import os

_sns_client = None  # noqa: F824


# improves speed for connecting on subsequent calls
def _get_sns_client():
    global _sns_client
    import boto3
    region_name = os.getenv('AWS_REGION')
    _sns_client = boto3.session.Session().client('sns', region_name=region_name) if _sns_client is None else _sns_client
    return _sns_client



from collections.abc import Iterable
from enum import Enum
from numpy import around
from numpy.typing import NDArray
from typing import Optional, Union


def calc_descriptive_stats(
    arr: NDArray,
    stats_definition: Union[Enum, str],
    axis: Optional[int] = None,
    decimals: int = 6
) -> dict:
    """
    Calculate the descriptive stats for an array row-wise, round them to specified number of decimal places and return
    them formatted for a HESTIA node.

    Parameters
    ----------
    arr : NDArray
    stats_definition : Enum | str
    axis : int | None
    decimals : int

    Returns
    -------
    float
        The precision of the sample mean estimated by the Monte Carlo model as a floating point value with the same
        units as the estimated mean.
    """
    value = around(arr.mean(axis=axis), decimals)
    sd = around(arr.std(axis=axis), decimals)
    min_ = around(arr.min(axis=axis), decimals)
    max_ = around(arr.max(axis=axis), decimals)

    observations = (
        [arr.shape[0]] * arr.shape[1] if axis == 0
        else [arr.shape[1]] * arr.shape[0] if axis == 1
        else [arr.size]
    )

    return {
        "value": list(value) if isinstance(value, Iterable) else [value],
        "sd": list(sd) if isinstance(sd, Iterable) else [sd],
        "min": list(min_) if isinstance(min_, Iterable) else [min_],
        "max": list(max_) if isinstance(max_, Iterable) else [max_],
        "statsDefinition": stats_definition.value if isinstance(stats_definition, Enum) else stats_definition,
        "observations": observations
    }



import time
from dateutil.parser import parse
from statistics import mean
from functools import reduce
from math import log10, floor
import numpy
from hestia_earth.schema import NodeType


def to_precision(number: float, digits: int = 3) -> float:
    return round(number, digits - int(floor(log10(abs(number)))) - 1)


def non_empty_value(value) -> bool:
    """
    Check if a value is empty.

    Parameters
    ----------
    value
        Either a string, a list, a number or None.

    Returns
    -------
    bool
        True if the value is not en empty string or an empty list.
    """
    return all([
        value != '',
        value is not None,
        not isinstance(value, list) or value != [],
        not isinstance(value, dict) or value != {}
    ])


def non_empty_list(values: list) -> list:
    """
    Filter list removing empty values.

    Parameters
    ----------
    values
        A list of values.

    Returns
    -------
    list
        List without empty values.
    """
    return list(filter(non_empty_value, values))


def is_node_of(node_type) -> bool:
    """
    Check wether node is of a certain HESTIA Type.

    Parameters
    ----------
    node_type
        The type to check for.
    node
        The node.

    Returns
    -------
    bool
        True if matches type.
    """
    return lambda node: isinstance(node, dict) and node.get('type', node.get('@type')) == node_type.value


def is_term(node: dict) -> bool:
    """
    Check wether node is a `Term`.

    Parameters
    ----------
    node
        The node.

    Returns
    -------
    bool
        True if it is a `Term`.
    """
    return is_node_of(NodeType.TERM)(node)


def current_time_ms():
    """
    Get the time in ms since EPOCH.

    Returns
    -------
    int
        Time in milliseconds.
    """
    return float(time.time() * 1000)


def safe_parse_float(value: str, default=0):
    """
    Parse a string into a float.

    Parameters
    ----------
    value
        The string value to parse.
    default
        The default value if parsing not possible.

    Returns
    -------
    float
        The value as float or default value.
    """
    try:
        value = float(value)
        return default if numpy.isnan(value) else value
    except Exception:
        return default


def safe_parse_date(date=None, default=None):
    """
    Parse a string into a date.

    Parameters
    ----------
    value
        The string value to parse.
    default
        The default value if parsing not possible.

    Returns
    -------
    datetime
        The value as datetime or default value.
    """
    try:
        return parse(str(date), fuzzy=True)
    except Exception:
        return default


def is_number(value):
    """
    Return `True` if the value is either an `int` or a `float`.
    """
    return all([
        not isinstance(value, bool),  # True is apparently considered an `int`
        any([isinstance(value, int), isinstance(value, float)])
    ])


def is_boolean(v):
    """
    Return `True` if the value is a `bool`.
    """
    return isinstance(v, bool)


def list_average(value: list, default=0):
    """
    Returns the average over a list of numbers.

    Parameters
    ----------
    value
        A list of numbers.
    default
        The default value if the value does not contain a list of numbers.

    Returns
    -------
    float
        The average of the values.
    """
    values = non_empty_list(value) if value and isinstance(value, list) and all(map(is_number, value)) else []
    return mean(values) if values else default


def list_sum(value: list, default=0):
    """
    Returns the sum over a list of numbers.

    Parameters
    ----------
    value
        A list of numbers.
    default
        The default value if the value does not contain a list of numbers.

    Returns
    -------
    float
        The sum of the values.
    """
    values = non_empty_list(value) if value and isinstance(value, list) and all(map(is_number, value)) else []
    return sum(values) if values else default


def flatten(values: list):
    """
    Flattens a two-dimensional list into a one-dimensional list.

    Parameters
    ----------
    values
        A list of list.

    Returns
    -------
    list
        A list of single values.
    """
    return list(reduce(lambda x, y: x + (y if isinstance(y, list) else [y]), values, []))


def _get_by_key(x, y):
    return x if x is None else (
        x.get(y) if isinstance(x, dict) else list(map(lambda v: get_dict_key(v, y), x))
    )


def get_dict_key(value: dict, key: str): return reduce(lambda x, y: _get_by_key(x, y), key.split('.'), value)



import pandas as pd

from .cycle import get_cycle_emissions_calculation_status


def _emissions_color(row):
    color = (
        'red' if row['emissions-missing'] > 0 else
        'yellow' if row['emissions-incomplete'] > 0 else
        'lightgreen'
    )
    return [f"background-color: {color}"] * len(row)


def _emissions_with_status(cycle: dict):
    emissions = get_cycle_emissions_calculation_status(cycle)
    all_emissions = emissions.keys()
    # an emission is missing if there is no value (ignore `missingInputs`)
    missing_emissions = set([
        k for k, v in emissions.items()
        if len((set(v.keys()) - set(['missingInputs']))) == 0
    ])
    # an emission is incomplete if it has missing inputs
    incomplete_emissions = set([
        k for k, v in emissions.items()
        if all([
            len(v.get('missingInputs', [])) > 0,
            k not in missing_emissions
        ])
    ])
    complete_emissions = set([
        k for k, v in emissions.items()
        if all([
            len(v.get('missingInputs', [])) == 0,
            len((set(v.keys()) - set(['missingInputs']))) > 0
        ])
    ])
    return {
        'emissions-total': len(all_emissions),
        'emissions-complete': len(complete_emissions),
        'emissions-incomplete': len(incomplete_emissions),
        'emissions-missing': len(missing_emissions),
        'emissions': emissions,
    }


def _handle_lists(df: pd.DataFrame, columns: list):
    for col in columns:
        df[col] = df[col].apply(lambda v: ';'.join(v) if isinstance(v, list) else v)
    return df


def get_nodes_calculations_status_dataframe(nodes: list, file_format: str = 'excel'):
    cycles_status = [
        {
            'id': cycle.get('@id') or cycle.get('id')
        } | _emissions_with_status(cycle)
        for cycle in nodes
        if (cycle.get('@type') or cycle.get('type')) == 'Cycle'
    ]
    df = pd.json_normalize(cycles_status, errors='ignore')
    # convert list of inputs to semi-column strings
    list_columns = [col for col in df.columns if col.endswith(".inputs") or col.endswith(".missingInputs")]
    df = _handle_lists(df, list_columns)
    return df.style.apply(_emissions_color, axis=1) if file_format == 'excel' else df



"""
Based on code by Cool Farm Tool: https://gitlab.com/MethodsCFT/coolfarm-soc/-/blob/main/src/cfasoc/builders.py
"""
import hashlib
from functools import reduce
from numpy import (
    cumsum, dot, full, linalg, hstack, random, mean, vstack, abs, array, concatenate, exp, float64, inf, pi, prod, sign,
    sqrt
)
from numpy.typing import NDArray, DTypeLike
from typing import Union


def repeat_single(shape: tuple, value: float, dtype: DTypeLike = None) -> NDArray:
    """
    Repeat a single value to form an array of a defined shape.

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    value : float
        Value to be repeated.
    dtype : DTypeLike, optional
        The desired data-type for the array.

    Returns
    -------
    NDArray
        Array with repeated value.
    """
    return full(shape=shape, fill_value=value, dtype=dtype)


def repeat_array_as_columns(n_iterations: int, arr: NDArray) -> NDArray:
    """
    Repeat a numpy array horizontally as columns.

    Parameters
    ----------
    n_iterations : int
        Number of times the columns should be repeated.
    arr : NDArray
        Array to repeat.

    Returns
    -------
    NDArray
        Repeated array.
    """
    return hstack([arr for _ in range(n_iterations)])


def repeat_array_as_rows(n_iterations: int, arr: NDArray) -> NDArray:
    """
    Repeat a numpy array vertically as rows.

    Parameters
    ----------
    n_iterations : int
        Number of times the rows should be repeated.
    arr : NDArray
        Array to repeat.

    Returns
    -------
    NDArray
        Repeated array.
    """
    return vstack([arr for _ in range(n_iterations)])


def repeat_1d_array_as_columns(n_columns: int, column: NDArray) -> NDArray:
    """
    Repeat a column (NDArray) to form an array of a defined shape

    Parameters
    ----------
    n_columns : int
        How many times the column (NDArray) should be repeated.
    column : NDArray
        The column (NDArray) to be repeated.

    Returns
    -------
    NDArray
        Repeated array.
    """
    return vstack([column for _ in range(n_columns)]).transpose()


def discrete_uniform_1d(
    shape: tuple, low: float, high: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a discrete uniform distribution and produce an array of a specified shape.
    All rows in a specified column will have the same sample value, but each column will be different (1 dimensional
    variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    low : float
        Lower bound of the discrete uniform distribution to be sampled.
    high : float
        Upper bound of the discrete uniform distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 1 dimensional variability.
    """
    n_rows, n_columns = shape
    rng = random.default_rng(seed)
    return repeat_array_as_rows(
        n_rows,
        rng.uniform(low=low, high=high, size=n_columns)
    )


def discrete_uniform_2d(
    shape: tuple, low: float, high: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a discrete uniform distribution and produce an array of a specified shape.
    All rows and columns contain different sample values (2 dimensional variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    low : float
        Lower bound of the discrete uniform distribution to be sampled.
    high : float
        Upper bound of the discrete uniform distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 2 dimensional variability.
    """
    rng = random.default_rng(seed)
    return rng.uniform(low=low, high=high, size=shape)


def triangular_1d(
    shape: tuple, low: float, high: float, mode: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a triangular distribution and produce an array of a specified shape.
    All rows in a specified column will have the same sample value, but each column will be different (1 dimensional
    variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    low : float
        Lower bound of the triangular distribution to be sampled.
    high : float
        Upper bound of the triangular distribution to be sampled.
    mode : float
        Mode of the triangular distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 1 dimensional variability.
    """
    n_rows, n_columns = shape
    rng = random.default_rng(seed)
    return repeat_array_as_rows(
        n_rows,
        rng.triangular(left=low, mode=mode, right=high, size=n_columns)
    )


def triangular_2d(
    shape: tuple, low: float, high: float, mode: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a triangular distribution and produce an array of a specified shape.
    All rows and columns contain different sample values (2 dimensional variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    low : float
        Lower bound of the triangular distribution to be sampled.
    high : float
        Upper bound of the triangular distribution to be sampled.
    mode : float
        Mode of the triangular distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 2 dimensional variability.
    """
    rng = random.default_rng(seed)
    return rng.triangular(left=low, mode=mode, right=high, size=shape)


def normal_1d(
    shape: tuple, mu: float, sigma: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a normal distribution and produce an array of a specified shape.
    All rows in a specified column will have the same sample value, but each column will be different (1 dimensional
    variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    mu : float
        Mean of the normal distribution to be sampled.
    sigma : float
        Standard deviation of the normal distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 1 dimensional variability.
    """
    n_rows, n_columns = shape
    rng = random.default_rng(seed)
    return repeat_array_as_rows(
        n_rows,
        rng.normal(loc=mu, scale=sigma, size=n_columns)
    )


def normal_2d(
    shape: tuple, mu: float, sigma: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a normal distribution and produce an array of a specified shape.
    All rows and columns contain different sample values (2 dimensional variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    mu : float
        Mean of the normal distribution to be sampled.
    sigma : float
        Standard deviation of the normal distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 2 dimensional variability.
    """
    rng = random.default_rng(seed)
    return rng.normal(loc=mu, scale=sigma, size=shape)


def truncated_normal_1d(
    shape: tuple, mu: float, sigma: float, low: float, high: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a truncated normal distribution and produce an array of a specified shape.
    All rows in a specified column will have the same sample value, but each column will be different (1 dimensional
    variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    mu : float
        Mean of the normal distribution to be sampled.
    sigma : float
        Standard deviation of the normal distribution to be sampled.
    low : float
        Lower bound of the normal distribution to be sampled.
    high : float
        Upper bound of the normal distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 1 dimensional variability.
    """
    n_rows, n_columns = shape
    return repeat_array_as_rows(
        n_rows,
        truncnorm_rvs(a=low, b=high, loc=mu, scale=sigma, shape=n_columns, seed=seed)
    )


def truncated_normal_2d(
    shape: tuple, mu: float, sigma: float, low: float, high: float, seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Sample from a truncated normal distribution and produce an array of a specified shape.
    All rows and columns contain different sample values (2 dimensional variability).

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    mu : float
        Mean of the normal distribution to be sampled.
    sigma : float
        Standard deviation of the normal distribution to be sampled.
    low : float
        Lower bound of the normal distribution to be sampled.
    high : float
        Upper bound of the normal distribution to be sampled.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 2 dimensional variability.
    """
    return truncnorm_rvs(a=low, b=high, loc=mu, scale=sigma, shape=shape, seed=seed)


def plus_minus_uncertainty_to_normal_1d(
    shape: tuple,
    value: float,
    uncertainty: float,
    confidence_interval: float = 95,
    seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Return a normally distributed sample given a value and uncertainty expressed as +/- a percentage.

    All rows in a specified column will have the same sample value, but each column will be different (1 dimensional
    variability).

    This function has been written to serve Table 5.5b on Page 5.32, Tier 2 Steady State Method for Mineral Soils,
    Chapter 5 Cropland, 2019 Refinement to the 2006 IPCC Guidelines for National Greenhouse Gas Inventories. Table 5.5b
    notes:

        "Uncertainty is assumed to be ±75% for the N content estimates and ±50% for the lignin content estimates,
        expressed as a 95% confidence intervals."

    This function also serves Table 11.2 on Page 11.19, Tier 2 Steady State Method for Mineral Soils, Chapter 11 N2O
    Emissions from Managed Soils, and CO2 Emissions from Lime and Urea Application, 2019 Refinement to the 2006 IPCC
    Guidelines for National Greenhouse Gas Inventories.

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    value : float
        Reported value.
    uncertainty : float
        Uncertainty expressed as +/- a percentage.
    confidence_interval : float
        Confidence interval the uncertainty represents.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 1 dimensional variability.
    """
    n_rows, n_columns = shape
    n_sds = calc_z_critical(confidence_interval)
    sigma = (value * (uncertainty / 100)) / n_sds
    return repeat_array_as_rows(
        n_rows,
        normal_1d(shape=(1, n_columns), mu=value, sigma=sigma, seed=seed)
    )


def plus_minus_uncertainty_to_normal_2d(
    shape: tuple,
    value: float,
    uncertainty: float,
    confidence_interval: float = 95,
    seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Return a normally distributed sample given a value and uncertainty expressed as +/- a percentage.

    All rows and columns contain different sample values (2 dimensional variability).

    This function has been written to serve Table 5.5b on Page 5.32, Tier 2 Steady State Method for Mineral Soils,
    Chapter 5 Cropland, 2019 Refinement to the 2006 IPCC Guidelines for National Greenhouse Gas Inventories. Table 5.5b
    notes:

        "Uncertainty is assumed to be ±75% for the N content estimates and ±50% for the lignin content estimates,
        expressed as a 95% confidence intervals."

    This function also serves Table 11.2 on Page 11.19, Tier 2 Steady State Method for Mineral Soils, Chapter 11 N2O
    Emissions from Managed Soils, and CO2 Emissions from Lime and Urea Application, 2019 Refinement to the 2006 IPCC
    Guidelines for National Greenhouse Gas Inventories.

    Parameters
    ----------
    shape : tuple
        Shape (rows, columns).
    value : float
        Reported value.
    uncertainty : float
        Uncertainty expressed as +/- a percentage.
    confidence_interval : float
        Confidence interval the uncertainty represents.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples with 2 dimensional variability.
    """
    n_sds = calc_z_critical(confidence_interval)
    sigma = (value * (uncertainty / 100)) / n_sds
    return normal_2d(shape=shape, mu=value, sigma=sigma, seed=seed)


def grouped_avg(arr: NDArray, n: int = 12) -> NDArray:
    """ Row-wise averaging of numpy arrays. For example:
    1   2   3
    4   5   6
    7   8   9
    10  11  12
    13  14  15
    16  17  18

    if n = 6, becomes:
    8.5 9.5 10.5

    because:
    (1 + 4 + 7 + 10 + 13 + 16) / 6 = 8.5
    (2 + 5 + 8 + 11 + 14 + 17) / 6 = 9.5
    etc.

    if n = 2, becomes:
    2.5  3.5  4.5
    8.5  9.5  10.5
    14.5 15.5 16.5

    because:
    (in column 0) (1 + 4) / 2 = 2.5, (7 + 10) / 2 = 8.5, (13 + 16) / 2 = 14.5
    (in column 1) (2 + 5) / 2 = 3.5, (8 + 11) / 2 = 9.5, (14 + 17) / 2 = 15.5

    Source: https://stackoverflow.com/questions/30379311/fast-way-to-take-average-of-every-n-rows-in-a-npy-array

    Parameters
    ----------
    arr : NDArray
        Input array.
    n : int, optional
        Number of rows to average. Defaults to 12.

    Returns
    -------
    NDArray
        Output array
    """
    result = cumsum(arr, 0)[n-1::n] / float(n)
    result[1:] = result[1:] - result[:-1]
    return result


def avg_run_in_columnwise(arr: NDArray, n: int):
    """
    Reduce the first `n` elements of each column in an array by averaging them, while leaving the rest of the array
    modified.

    Parameters
    ----------
    arr : NDArray
        Input array.
    n : int
        The number of run-in elements to average.

    Returns
    -------
    NDArray
        The new array where the first element in each column is an average of the run in elements.
    """
    run_in: NDArray = mean(arr[:n], 0)
    return vstack([run_in, arr[n:]])


def avg_run_in_rowwise(arr: NDArray, n: int):
    """
    Reduce the first `n` elements of each row in an array by averaging them, while leaving the rest of the array
    modified.

    Parameters
    ----------
    arr : NDArray
        Input array.
    n : int
        The number of run-in elements to average.

    Returns
    -------
    NDArray
        The new array where the first element in each row is an average of the run in elements.
    """
    return avg_run_in_columnwise(arr.transpose(), n).transpose()


def gen_seed(node: dict, *args: tuple[str]) -> int:
    """
    Generate a seed based on a node's `@id` and optional args so that rng is the same each time the model is re-run.
    """
    node_id = node.get("@id", "")
    seed_str = "".join([node_id] + [str(arg) for arg in args])
    hashed = hashlib.shake_128(seed_str.encode(), usedforsecurity=False).hexdigest(4)
    return abs(int(hashed, 16))


def correlated_normal_2d(
    n_iterations: int,
    means: NDArray,
    sds: NDArray,
    correlation_matrix: NDArray,
    seed: Union[int, random.Generator, None] = None,
) -> NDArray:
    """
    Generate correlated random samples from a multivariate normal distribution with specified means, standard
    deviations, and a correlation matrix. Each row represents a different variable (e.g., different years), and each
    column represents a different iteration (sample).

    Parameters
    ----------
    n_iterations : int
        The number of samples (iterations) to generate for each variable.
    means : NDArray
        An array of mean values for each variable (row).
    sds : NDArray
        An array of standard deviations for each variable (row).
    correlation_matrix : NDArray
        A positive-definite matrix representing the correlations between the variables (rows).
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        A 2D array of shape (len(means), n_iterations), where each row corresponds to a different variable and each
        column corresponds to a sample iteration. The values in each row are correlated according to the provided
        correlation matrix.
    """
    # Generate independent random samples for each year
    shape = (len(means), n_iterations)
    independent_samples = normal_2d(shape, 0, 1, seed=seed)

    # Apply Cholesky decomposition to the correlation matrix
    cholesky_decomp = linalg.cholesky(correlation_matrix)

    # Apply Cholesky transformation to introduce correlation across years (rows) for each sample
    correlated_samples = dot(cholesky_decomp, independent_samples)

    # Scale by standard deviations and shift by means
    scaled_samples = (
        correlated_samples
        * repeat_1d_array_as_columns(n_iterations, sds)
        + repeat_1d_array_as_columns(n_iterations, means)
    )

    return scaled_samples


def calc_z_critical(
    confidence_interval: float,
    n_sided: int = 2
) -> float64:
    """
    Calculate the z-critical value from the confidence interval.

    Parameters
    ----------
    confidence_interval : float
        The confidence interval as a percentage between 0 and 100%.
    n_sided : int, optional
        The number of tails (default value = `2`).

    Returns
    -------
    float64
        The z-critical value as a floating point between 0 and infinity.
    """
    alpha = 1 - confidence_interval / 100
    return _normal_ppf(1 - alpha / n_sided)


def _normal_ppf(q: float64, tol: float64 = 1e-10) -> float64:
    """
    Calculates the percent point function (PPF), also known as the inverse cumulative distribution function (CDF), of a
    standard normal distribution using the Newton-Raphson method.

    Parameters
    ----------
    q : float64
        The quantile at which to evaluate the PPF.
    tol : float64, optional
        The tolerance for the Newton-Raphson method. Defaults to 1e-10.

    Returns
    -------
    float64
        The PPF value at the given quantile.
    """
    INITIAL_GUESS = 0
    MAX_ITER = 100

    def step(x):
        """Perform one step of the Newton-Raphson method."""
        x_new = x - (_normal_cdf(x) - q) / _normal_pdf(x)
        return x_new if abs(x_new - x) >= tol else x

    return (
        inf if q == 1 else
        -inf if q == 0 else
        reduce(lambda x, _: step(x), range(MAX_ITER), INITIAL_GUESS)
    )


def _normal_cdf(x: float64) -> float64:
    """
    Calculates the cumulative distribution function (CDF) of a standard normal distribution for a single value using a
    custom error function (erf).

    Parameters
    ----------
    x : float64
        The point at which to evaluate the CDF.

    Returns
    -------
    float64
        The CDF value at the given point.
    """
    return 0.5 * (1 + _erf(x / sqrt(2)))


def _erf(x: float64) -> float64:
    """
    Approximates the error function of a standard normal distribution using a numerical approximation based on
    Abramowitz and Stegun formula 7.1.26.

    Parameters
    ----------
    x : float64
        The input value.

    Returns
    -------
    float64
        The approximated value of the error function.
    """
    # constants
    A_1 = 0.254829592
    A_2 = -0.284496736
    A_3 = 1.421413741
    A_4 = -1.453152027
    A_5 = 1.061405429
    P = 0.3275911

    # Save the sign of x
    sign_ = sign(x)
    x_ = abs(x)

    # A&S formula 7.1.26
    t = 1.0 / (1.0 + P * x_)
    y = 1.0 - (((((A_5 * t + A_4) * t) + A_3) * t + A_2) * t + A_1) * t * exp(-x_ * x_)

    return sign_ * y


def _normal_pdf(x: float64) -> float64:
    """
    Calculates the probability density function (PDF) of a standard normal distribution for a single value.

    Parameters
    ----------
    x : float64
        The point at which to evaluate the PDF.

    Returns
    -------
    float64
        The PDF value at the given point.
    """
    return 1 / sqrt(2 * pi) * exp(-0.5 * x**2)


def _calc_confidence_level(
    z_critical: float64,
    n_sided: int = 2
) -> float64:
    """
    Calculate the confidence interval from the z-critical value.

    Parameters
    ----------
    z_critical_value : np.float64
        The confidence interval as a floating point number between 0 and infinity.
    n_sided : int, optional
        The number of tails (default value = `2`).

    Returns
    -------
    np.float64
        The confidence interval as a percentage between 0 and 100%.
    """
    alpha = (1 - _normal_cdf(z_critical)) * n_sided
    return (1 - alpha) * 100


def calc_required_iterations_monte_carlo(
    confidence_level: float,
    precision: float,
    sd: float
) -> int:
    """
    Calculate the number of iterations required for a Monte Carlo simulation to have a desired precision, subject to a
    given confidence level.

    Parameters
    ----------
    confidence_level : float
        The confidence level, as a percentage out of 100, that the precision should be subject too (i.e., we are x%
        sure that the sample mean deviates from the true populatation mean by less than the desired precision).
    precision : float
        The desired precision as a floating point value (i.e., if the Monte Carlo simulation will be used to estimate
        `organicCarbonPerHa` to a precision of 100 kg C ha-1 this value should be 100).
    sd : float
        The standard deviation of the sample. This can be estimated by running the model 500 times (a number that does
        not take too much time to run but is large enough for the sample standard deviation to converge reasonably
        well).

    Returns
    -------
    int
        The required number of iterations.
    """
    z_critical_value = calc_z_critical(confidence_level)
    return round(((sd * z_critical_value) / precision) ** 2)


def calc_confidence_level_monte_carlo(
    n_iterations: int,
    precision: float,
    sd: float
) -> float:
    """
    Calculate the confidence level that the sample mean calculated by the Monte Carlo simulation deviates from the
    true population mean by less than the desired precision.

    Parameters
    ----------
    n_iterations : int
        The number of iterations that the Monte Carlo simulation was run for.
    precision : float
        The desired precision as a floating point value (i.e., if the Monte Carlo simulation will be used to estimate
        `organicCarbonPerHa` to a precision of 100 kg C ha-1 this value should be 100).
    sd : float
        The standard deviation of the sample.

    Returns
    -------
    float
        The confidence level, as a percentage out of 100, that the precision should be subject too (i.e., we are x%
        sure that the sample mean deviates from the true populatation mean by less than the desired precision).
    """
    return _calc_confidence_level(precision*sqrt(n_iterations)/sd)


def calc_precision_monte_carlo(
    confidence_level: float,
    n_iterations: int,
    sd: float
) -> float:
    """
    Calculate the +/- precision of a Monte Carlo simulation for a desired confidence level.

    Parameters
    ----------
    confidence_level : float
        The confidence level, as a percentage out of 100, that the precision should be subject too (i.e., we are x%
        sure that the sample mean deviates from the true populatation mean by less than the desired precision).
    n_iterations : int
        The number of iterations that the Monte Carlo simulation was run for.
    sd : float
        The standard deviation of the sample.

    Returns
    -------
    float
        The precision of the sample mean estimated by the Monte Carlo model as a floating point value with the same
        units as the estimated mean.
    """
    z_critical = calc_z_critical(confidence_level)
    return (sd*z_critical)/sqrt(n_iterations)


def truncnorm_rvs(
    a: float,
    b: float,
    loc: float,
    scale: float,
    shape: Union[int, tuple[int, ...]],
    seed: Union[int, random.Generator, None] = None
) -> NDArray:
    """
    Generate random samples from a truncated normal distribution. Unlike the `scipy` equivalent, the `a` and `b` values
    are the abscissae at which we wish to truncate the distribution (as opposed to the number of standard deviations
    from `loc`).

    Parameters
    ----------
    a : float
        The lower bound of the distribution.
    b : float
        The upper bound of the distribution.
    loc : float
        Mean ("centre") of the distribution.
    scale : float
        Standard deviation (spread or "width") of the distribution. Must be non-negative.
    size : int | tuple[int, ...]
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k samples are drawn.
    seed : int | Generator | None, optional
        A seed to initialize the BitGenerator. If passed a Generator, it will be returned unaltered. If `None`, then
        fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    NDArray
        Array of samples.
    """
    size = prod(shape)
    samples = array([])
    rng = random.default_rng(seed)

    while samples.size < size:
        samples_temp = rng.normal(loc, scale, (size - samples.size) * 2)
        valid_samples = samples_temp[(a <= samples_temp) & (samples_temp <= b)]
        samples = concatenate([samples, valid_samples])

    return samples[:size].reshape(shape)


def add_normal_distributions(
    mu_1: float, sigma_1: float, mu_2: float, sigma_2: float, rho: float = 0
) -> tuple[float, float]:
    """
    Add together two normal distributions, with optional correlation.

    Given two normal distributions **X<sub>1</sub> ~ N(mu<sub>1</sub>, sigma<sub>1</sub><sup>2</sup>)** and
    **X<sub>2</sub> ~ N(mu<sub>2</sub>, sigma<sub>2</sub><sup>2</sup>)**, this function calculates the resulting mean
    and standard deviation of the sum **Z = X<sub>1</sub> + X<sub>2</sub>**, taking into account the correlation
    between them.

    n.b. Positive correlations (`rho` > `0`) increase the standard deviation of **Z** because positively correlated
    variables tend to move together, increasing combined uncertainty. Negative correlations (`rho` < `0`) reduces the
    standard deviation since the variables move in opposite directions, cancelling out some of the variability.
    Independant variables (`rho` = `0`) result in an intermediate level of uncertainty.

    Parameters
    ----------
    mu_1 : float
        Mean of the first normal distribution (X<sub>1</sub>).
    sigma_1 : float
        Standard deviation of the first normal distribution (X<sub>1</sub>).
    mu_2 : float
        Mean of the second normal distribution (X<sub>2</sub>).
    sigma_2 : float
        Standard deviation of the second normal distribution (X<sub>2</sub>).
    rho : float, optional
        Correlation coefficient between **X<sub>1</sub>** and **X<sub>2</sub>**. `rho` must be a value between -1
        (perfectly negative correlation) and 1 (perfectly positive correlation). Default is 0 (independent variables).

    Returns
    -------
    tuple[float, float]
        A tuple in the shape `(mu_sum, sigma_sum)` containing the mean and standard deviation of the distribution
        **Z = X<sub>1</sub> + X<sub>2</sub>**.
    """
    mu_sum = mu_1 + mu_2
    sigma_sum = sqrt(
        sigma_1 ** 2
        + sigma_2 ** 2
        + 2 * rho * sigma_1 * sigma_2
    )
    return mu_sum, sigma_sum


def subtract_normal_distributions(
    mu_1: float, sigma_1: float, mu_2: float, sigma_2: float, rho: float = 0
) -> tuple[float, float]:
    """
    Subtract a normal distribution from another, with optional correlation.

    Given two normal distributions **X<sub>1</sub> ~ N(mu<sub>1</sub>, sigma<sub>1</sub><sup>2</sup>)** and
    **X<sub>2</sub> ~ N(mu<sub>2</sub>, sigma<sub>2</sub><sup>2</sup>)**, this function calculates the resulting mean
    and standard deviation of the difference **Z = X<sub>1</sub> - X<sub>2</sub>**, taking into account the correlation
    between them.

    n.b. Positive correlations (`rho` > `0`) reduce the standard deviation of **Z** because positively correlated
    variables tend to move together, cancelling out some of the variability when subtracted. Negative correlations
    (`rho` < `0`) increase the standard deviation since the variables move in opposite directions, amplifying the
    variability when subtracted. Independant variables (`rho` = `0`) result in an intermediate level of uncertainty.

    Parameters
    ----------
    mu_1 : float
        Mean of the first normal distribution (X<sub>1</sub>).
    sigma_1 : float
        Standard deviation of the first normal distribution (X<sub>1</sub>).
    mu_2 : float
        Mean of the second normal distribution (X<sub>2</sub>).
    sigma_2 : float
        Standard deviation of the second normal distribution (X<sub>2</sub>).
    rho : float, optional
        Correlation coefficient between **X<sub>1</sub>** and **X<sub>2</sub>**. `rho` must be a value between -1
        (perfectly negative correlation) and 1 (perfectly positive correlation). Default is 0 (independent variables).

    Returns
    -------
    tuple[float, float]
        A tuple in the shape `(mu_diff, sigma_diff)` containing the mean and standard deviation of the distribution
        **Z = X<sub>1</sub> - X<sub>2</sub>**.
    """
    mu_sum = mu_1 - mu_2
    sigma_sum = sqrt(
        sigma_1 ** 2
        + sigma_2 ** 2
        - 2 * rho * sigma_1 * sigma_2
    )
    return mu_sum, sigma_sum


def lerp_normal_distributions(
    mu_1: float,
    sigma_1: float,
    mu_2: float,
    sigma_2: float,
    alpha: float,
    rho: float = 0
) -> tuple[float, float]:
    """
    Linearly interpolate between two normal distributions, with optional correlation.

    Given two normal distributions **X<sub>1</sub> ~ N(mu<sub>1</sub>, sigma<sub>1</sub><sup>2</sup>)** and
    **X<sub>2</sub> ~ N(mu<sub>2</sub>, sigma<sub>2</sub><sup>2</sup>)**, this function calculates the resulting mean
    and standard deviation of the interpolated distribution **Z = (1 - alpha) * X<sub>1</sub> + alpha * X<sub>2</sub>**,
    taking into account the correlation between them.

    n.b. Positive correlations (`rho` > `0`) increase the standard deviation of **Z** because positively correlated
    variables tend to move together, increasing combined uncertainty. Negative correlations (`rho` < `0`) reduces the
    standard deviation since the variables move in opposite directions, cancelling out some of the variability.
    Independant variables (`rho` = `0`) result in an intermediate level of uncertainty.

    Parameters
    ----------
    mu_1 : float
        Mean of the first normal distribution (X<sub>1</sub>).
    sigma_1 : float
        Standard deviation of the first normal distribution (X<sub>1</sub>).
    mu_2 : float
        Mean of the second normal distribution (X<sub>2</sub>).
    sigma_2 : float
        Standard deviation of the second normal distribution (X<sub>2</sub>).
    alpha : float
        Interpolation factor (0 <= alpha <= 1). A value of 0 results in X1, a value of 1 results in X2, and values
        between 0 and 1 interpolate between the two. Values of below 0 and above 1 will extrapolate beyond the
        X<sub>1</sub> and X<sub>2</sub> respectively.
    rho : float, optional
        Correlation coefficient between **X<sub>1</sub>** and **X<sub>2</sub>**. `rho` must be a value between -1
        (perfectly negative correlation) and 1 (perfectly positive correlation). Default is 0 (independent variables).

    Returns
    -------
    tuple[float, float]
        A tuple in the shape `(mu_Z sigma_Z)` containing the mean and standard deviation of the distribution
        **Z = (1 - alpha) * X<sub>1</sub> + alpha * X<sub>2</sub>**.
    """
    mu_Z = (1 - alpha) * mu_1 + alpha * mu_2
    var_Z = (
        ((1 - alpha) ** 2) * sigma_1 ** 2
        + (alpha ** 2) * sigma_2 ** 2
        + 2 * alpha * (1 - alpha) * rho * sigma_1 * sigma_2
    )
    sigma_Z = sqrt(var_Z)
    return mu_Z, sigma_Z



from typing import List
from hestia_earth.schema import TermTermType

from .lookup import download_lookup, lookup_term_ids
from .lookup_utils import (
    is_in_system_boundary,
    is_siteType_allowed,
    is_site_measurement_id_allowed,
    is_product_termType_allowed,
    is_product_id_allowed,
    is_input_termType_allowed,
    is_input_id_allowed
)


def emissions_in_system_boundary(termType: TermTermType = TermTermType.EMISSION) -> List[str]:
    """
    Get all emissions included in HESTIA system boundary.

    Returns
    -------
    List[str]
        List of emission IDs
    """
    lookup = download_lookup(f"{termType.value}.csv")
    # find all emissions in system boundary
    return list(filter(is_in_system_boundary, lookup_term_ids(lookup)))


def cycle_emission_is_in_system_boundary(cycle: dict):
    def filter_term(term_id: str):
        return is_in_system_boundary(term_id) and all(map(lambda check: check(cycle, term_id), [
            is_siteType_allowed,
            is_site_measurement_id_allowed,
            is_product_termType_allowed,
            is_product_id_allowed,
            is_input_termType_allowed,
            is_input_id_allowed
        ]))

    return filter_term


def cycle_emissions_in_system_boundary(cycle: dict, termType: TermTermType = TermTermType.EMISSION):
    """
    Get all emissions relevant for the Cycle, included in HESTIA system boundary.

    Returns
    -------
    List[str]
        List of emission IDs
    """
    lookup = download_lookup(f"{termType.value}.csv")
    # find all emissions in system boundary
    return list(filter(cycle_emission_is_in_system_boundary(cycle), lookup_term_ids(lookup)))



import json
import numpy as np


EXCLUDE_FIELDS = ["@type", "type", "@context"]
EXCLUDE_PRIVATE_FIELDS = [
    "added", "addedVersion",
    "updated", "updatedVersion",
    "aggregatedVersion",
    "_cache"
]


def _with_csv_formatting(dct):
    """
    Use as object_hook when parsing a JSON node: json.loads(node, object_hook=_with_csv_formatting).
    Ensures parsed JSON has field values formatted according to hestia csv conventions.
    """
    if 'boundary' in dct:
        dct['boundary'] = json.dumps(dct['boundary'])
    for key, value in dct.items():
        if _is_scalar_list(value):
            dct[key] = ';'.join([str(el) for el in value])
    return dct


def _is_scalar_list(value):
    if not isinstance(value, list):
        return False
    all_scalar = True
    for element in value:
        if not np.isscalar(element):
            all_scalar = False
            break
    return all_scalar



import os
import pandas as pd
from unittest.mock import patch, call

from tests.utils import fixtures_path
from hestia_earth.utils.pivot.pivot_csv import pivot_csv, pivot_hestia_file

class_path = 'hestia_earth.utils.pivot.pivot_csv'
fixtures_folder = os.path.join(fixtures_path, 'pivot', 'pivot_csv')


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={
        "Full tillage": "fullTillage",
        "Diesel": "diesel",
        "Inorganic Potassium fertiliser, unspecified (kg K2O)": "inorganicPotassiumFertiliserUnspecifiedKgK2O",
        "Inorganic Phosphorus fertiliser, unspecified (kg P2O5)": "inorganicPhosphorusFertiliserUnspecifiedKgP2O5",
        "Urea (kg N)": "ureaKgN",
        "Peanut, in shell": "peanutInShell",
    },
)
def test_pivot_csv_cycle(mock):
    filepath = f"{fixtures_folder}/cycle.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/cycle-pivoted.csv", index_col=None, dtype=object
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()
    mock.assert_has_calls(
        [
            call(
                [
                    "Diesel",
                    "Full tillage",
                    "Inorganic Phosphorus fertiliser, unspecified (kg P2O5)",
                    "Inorganic Potassium fertiliser, unspecified (kg K2O)",
                    "Peanut, in shell",
                    "Urea (kg N)",
                ]
            )
        ]
    )


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={
        "Eutrophication potential, excluding fate": "eutrophicationPotentialExcludingFate",
        "GWP100": "gwp100",
        "N2O, to air, organic fertiliser, direct": "n2OToAirOrganicFertiliserDirect",
        "N2O, to air, inorganic fertiliser, direct": "n2OToAirInorganicFertiliserDirect",
    },
)
def test_pivot_csv_impact(mock):
    filepath = f"{fixtures_folder}/impact.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/impact-pivoted.csv", index_col=None, dtype=object
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()
    mock.assert_has_calls(
        [
            call(
                [
                    "Eutrophication potential, excluding fate",
                    "GWP100",
                    "N2O, to air, inorganic fertiliser, direct",
                    "N2O, to air, organic fertiliser, direct",
                ]
            )
        ]
    )


def test_pivot_csv_multinode_rows():
    filepath = f"{fixtures_folder}/multinode-rows.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/multinode-rows-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={"Urea (kg N)": "ureaKgN"},
)
def test_pivot_csv_cycle_missing_ids(mock):
    filepath = f"{fixtures_folder}/missing-ids.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/missing-ids-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()
    mock.assert_has_calls([call(["Urea (kg N)"])])


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={"Irrigated": "irrigated"},
)
def test_pivot_csv_empty_cells(mock):
    filepath = f"{fixtures_folder}/empty-cells.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/empty-cells-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


def test_pivot_csv_preserves_uniqueness_fields():
    filepath = f"{fixtures_folder}/uniqueness-fields-undifferentiating.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/uniqueness-fields-undifferentiating-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={
        "Helicopter use, operation unspecified": "helicopterUseOperationUnspecified",
        "Cooling, with evaporative cooling tower": "coolingWithEvaporativeCoolingTower",
        "Small tractor use, operation unspecified": "smallTractorUseOperationUnspecified",
        "Coating seeds": "coatingSeeds",
        "Buttage of vine": "buttageOfVine",
    },
)
def test_pivot_csv_uniqueness_fields_differentiating(mock):
    filepath = f"{fixtures_folder}/uniqueness-fields-differentiating.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/uniqueness-fields-differentiating-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()
    mock.assert_has_calls(
        [
            call(
                [
                    "Buttage of vine",
                    "Coating seeds",
                    "Cooling, with evaporative cooling tower",
                    "Helicopter use, operation unspecified",
                    "Small tractor use, operation unspecified",
                ]
            )
        ]
    )


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={
        "Cooling, with evaporative cooling tower": "coolingWithEvaporativeCoolingTower",
    },
)
def test_pivot_csv_uniqueness_fields_non_matching(mock):
    filepath = f"{fixtures_folder}/uniqueness-fields-non-matching.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/uniqueness-fields-non-matching-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()
    mock.assert_has_calls([call(["Cooling, with evaporative cooling tower"])])


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={
        "Nitrogen content": "nitrogenContent",
    },
)
def test_pivot_csv_properties(mock):
    filepath = f"{fixtures_folder}/properties-exception.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/properties-exception-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()
    mock.assert_has_calls([call(["Nitrogen content"])])


def test_pivot_csv_depth():
    filepath = f"{fixtures_folder}/depth-exception.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/depth-exception-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


def test_pivot_csv_shuffled():
    filepath = f"{fixtures_folder}/shuffled.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/shuffled-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={"Full tillage": "fullTillage", "Urea (kg N)": "ureaKgN"},
)
def test_pivot_csv_cycle_deep(*args):
    filepath = f"{fixtures_folder}/deep.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/deep-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


def test_pivot_csv_non_node_arrayfields(*args):
    filepath = f"{fixtures_folder}/non-node-arrayfields.csv"
    expected = pd.read_csv(
        f"{fixtures_folder}/non-node-arrayfields-pivoted.csv",
        index_col=None,
        dtype=object,
    )
    df = pivot_csv(filepath)
    assert df.to_csv() == expected.to_csv()


@patch(
    f"{class_path}.find_term_ids_by_names",
    return_value={
        "Grinding, with grinder": "grinding",
        "Motor gasoline": "motorGasoline",
        "Orchard density": "orchardDensity",
    },
)
def test_pivot_hestia_file(*args):
    filepath = f"{fixtures_folder}/nodes.hestia"
    expected = pd.read_csv(
        f"{fixtures_folder}/nodes.hestia-pivoted.csv",
        index_col=None,
        dtype=object,
    )

    with open(filepath) as fd:
        hestia_file = fd.read()

    df = pivot_hestia_file(hestia_file)
    assert df.to_csv() == expected.to_csv()



import os
import json
import re
import numpy as np
import pandas as pd

from tests.utils import fixtures_path
from hestia_earth.utils.pivot.pivot_json import (
    _with_csv_formatting,
    pivot_nodes,
    pivot_hestia_file,
)
from flatten_json import unflatten_list
from hestia_earth.schema.utils.sort import SORT_CONFIG
from hestia_earth import schema

class_path = 'hestia_earth.utils.pivot.pivot_csv'
fixtures_folder = os.path.join(fixtures_path, 'pivot', 'pivot_json')

node_types = {k: getattr(schema, k)().fields for k in schema.SCHEMA_TYPES}
name_to_ids_mapping = {
    "Full tillage": "fullTillage",
    "Diesel": "diesel",
    "Motor gasoline": "motorGasoline",
    "Inorganic Potassium fertiliser, unspecified (kg K2O)": "inorganicPotassiumFertiliserUnspecifiedKgK2O",
    "Inorganic Phosphorus fertiliser, unspecified (kg P2O5)": "inorganicPhosphorusFertiliserUnspecifiedKgP2O5",
    "Urea (kg N)": "ureaKgN",
    "Peanut, in shell": "peanutInShell",
    "Eutrophication potential, excluding fate": "eutrophicationPotentialExcludingFate",
    "GWP100": "gwp100",
    "N2O, to air, organic fertiliser, direct": "n2OToAirOrganicFertiliserDirect",
    "N2O, to air, inorganic fertiliser, direct": "n2OToAirInorganicFertiliserDirect",
    "Irrigated": "irrigated",
    "Helicopter use, operation unspecified": "helicopterUseOperationUnspecified",
    "Cooling, with evaporative cooling tower": "coolingWithEvaporativeCoolingTower",
    "Small tractor use, operation unspecified": "smallTractorUseOperationUnspecified",
    "Coating seeds": "coatingSeeds",
    "Buttage of vine": "buttageOfVine",
    "Nitrogen content": "nitrogenContent",
    "Grinding, with grinder": "grinding",
    "Orchard density": "orchardDensity",
}


def _get_node_type(col):
    label = col.split(".")[0]
    return label[0].upper() + label[1:]


def _add_missing_fields(row, is_input, col, parent_type, prefix=""):
    subnode_col = re.search(r"(.+?\.\d+)\.(.+)", col)
    if not subnode_col:
        return None
    sub_node, deep_col = subnode_col.groups()
    node_type = (
        # We are not handling fields like subnode_type_A.subnode_type_B.0
        # We are always fetching type_A in this scenario.
        SORT_CONFIG.get(parent_type)
        .get(sub_node.split(".")[0])
        .get("type")
    )
    next_prefix = ".".join([el for el in (prefix, sub_node) if el])
    row[f"{next_prefix}.@type"] = node_type
    _add_missing_fields(row, is_input, deep_col, node_type, prefix=next_prefix)


def _row_to_dict(row, is_input, parent_type):
    row.dropna(inplace=True)
    if is_input:
        for col in row.index:
            _add_missing_fields(row, is_input, col, parent_type)
    return row.to_dict()


def _df_to_dict(df, is_input):
    df.index = map(lambda col: ".".join(col.split(".")[1:]), df.index)
    df.loc["@type"] = df.name
    dicts = df.apply(_row_to_dict, is_input=is_input, parent_type=df.name)
    return dicts


def _ensure_id_cols(df, name_to_ids):
    names_df = df.filter(regex=r"\.name", axis=1)
    for name_col in names_df.columns:
        id_col = name_col.replace(".name", ".@id")
        for idx, name in df[name_col].items():
            if id_col not in df:
                df[id_col] = np.nan
            if pd.isna(df.loc[idx, id_col]):
                df.loc[idx, id_col] = name_to_ids[name]


def _convert_csv_to_nodes(fixture, is_input, name_to_ids):
    """
    Gets json fixtures or creates them from corresponding csv files.
    Conversion for *-pivoted files is not perfect as we do not detect
    the difference between an empty cell which should be discarded
    (ie. header not used by a row) and a node without a value key
    (the latter are represented in csv as field.nodeId.value = None)
    """
    filepath = (
        f"{fixtures_path}/pivot/pivot_csv/{fixture}.csv"
        if is_input
        else f"{fixtures_path}/pivot/pivot_csv/{fixture}-pivoted.csv"
    )
    df = pd.read_csv(filepath, index_col=None, dtype=object)
    df.drop(columns="-", errors="ignore", inplace=True)
    df.replace("-", np.nan, inplace=True)
    df.replace(
        ["TRUE", "True", "true", "FALSE", "False", "false"],
        [True, True, True, False, False, False],
        inplace=True,
    )
    if is_input:
        df.dropna(how="all", axis=1, inplace=True)
    df.rename(lambda col: col.replace(".id", ".@id"), axis=1, inplace=True)
    if is_input:
        _ensure_id_cols(df, name_to_ids)
    df = df.T.groupby(_get_node_type).apply(_df_to_dict, is_input)
    nodes = [
        node for _node_type, nodes in df.iterrows() for node in nodes if node.get("@id")
    ]
    return nodes


def get_nodes_from_fixture(fixture, name_to_ids={}):
    try:
        with open(f"{fixtures_folder}/{fixture}.json") as file:
            input = json.load(file, object_hook=_with_csv_formatting)["nodes"]
        with open(f"{fixtures_folder}/{fixture}-pivoted.json") as file:
            expected = json.load(file, object_hook=_with_csv_formatting)["nodes"]
    except FileNotFoundError:
        print(f"\n{fixture} not found: attempting to create from csv.\n")
        name_to_ids.update({np.nan: np.nan})
        input = _convert_csv_to_nodes(fixture, True, name_to_ids)
        expected = _convert_csv_to_nodes(fixture, False, name_to_ids)

        input, expected = (
            [unflatten_list(node, ".") for node in input],
            [unflatten_list(node, ".") for node in expected],
        )
        with open(f"{fixtures_folder}/{fixture}.json", "w") as file:
            file.write(json.dumps({"nodes": input}, sort_keys=True, indent=2))
        with open(
            f"{fixtures_folder}/{fixture}-pivoted.json", "w"
        ) as file:
            file.write(json.dumps({"nodes": expected}, sort_keys=True, indent=2))

    return (input, expected)


def test_pivot_json_cycle():
    input, expected = get_nodes_from_fixture("cycle", name_to_ids_mapping)
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_impact():
    input, expected = get_nodes_from_fixture("impact", name_to_ids_mapping)
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_multinode_rows():
    input, expected = get_nodes_from_fixture("multinode-rows")
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_preserves_uniqueness_fields():
    input, expected = get_nodes_from_fixture(
        "uniqueness-fields-undifferentiating", name_to_ids_mapping
    )
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_uniqueness_fields_differentiating():
    input, expected = get_nodes_from_fixture(
        "uniqueness-fields-differentiating", name_to_ids_mapping
    )
    actual = pivot_nodes(input)
    assert expected == actual


# Output differs from CSV pivoter (see https://gitlab.com/hestia-earth/hestia-utils/-/issues/32)
def test_pivot_json_uniqueness_fields_non_matching():
    input, expected = get_nodes_from_fixture("uniqueness-fields-non-matching", name_to_ids_mapping)
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_properties():
    input, expected = get_nodes_from_fixture("properties-exception", name_to_ids_mapping)
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_depth():
    input, expected = get_nodes_from_fixture("depth-exception")
    actual = pivot_nodes(input)
    assert expected == actual


# Output differs from CSV pivoter (see https://gitlab.com/hestia-earth/hestia-utils/-/issues/32)
def test_pivot_json_cycle_deep():
    input, expected = get_nodes_from_fixture("deep", name_to_ids_mapping)
    actual = pivot_nodes(input)
    assert expected == actual


def test_pivot_json_node_arrayfields_merged():
    input, expected = get_nodes_from_fixture("node-arrayfields-merged")
    actual = pivot_nodes(input)

    assert expected == actual


def test_pivot_json_unindexed_node():
    input, expected = get_nodes_from_fixture("unindexed-node")
    actual = pivot_nodes(input)

    assert expected == actual


def test_pivot_hestia_file():
    _input, expected = get_nodes_from_fixture("nodes.hestia", name_to_ids_mapping)
    actual = pivot_hestia_file(
        open(f"{fixtures_folder}/nodes.hestia.json", "r").read()
    )
    assert expected == actual
